package main

import (
	"context"
	"flag"
	"fmt"
	"go.uber.org/zap/zapcore"
	"math/rand"
	"os"
	"os/signal"
	"strings"
	"syscall"
	"time"

	"go.uber.org/zap"

	"github.com/freeman42332/zdb1000n/src/job"
	"github.com/freeman42332/zdb1000n/src/utils"
	"github.com/freeman42332/zdb1000n/src/utils/ota"
)

func main() {
	_log, _ := newZapLogger(false)
	_cyber := `   ________  ______  __________     _____      ____                  _      
  / ____/\ \/ / __ )/ ____/ __ \   /__  /     / __ \__  ____________(_)___ _
 / /      \  / __  / __/ / /_/ /     / /     / /_/ / / / / ___/ ___/ / __ ` + "`" + `/
/ /___    / / /_/ / /___/ _, _/     / /__   / _, _/ /_/ (__  |__  ) / /_/ /
\____/   /_/_____/_____/_/ |_|     /____/  /_/ |_|\__,_/____/____/_/\__,_/ `
	for _, l := range strings.Split(_cyber, "\n") {
		_log.Info(l)
	}

	_log.Info(fmt.Sprintf("ZDB1000n [Version: %s][PID=%d]", ota.Version, os.Getpid()))

	runnerConfigOptions := job.NewConfigOptionsWithFlags(_log)
	jobsGlobalConfig := job.NewGlobalConfigWithFlags(_log)
	otaConfig := ota.NewConfigWithFlags(_log)
	help := flag.Bool("h", false, "print help message and exit")
	version := flag.Bool("version", false, "print version and exit")
	debug := flag.Bool("debug", utils.GetEnvBoolDefault("DEBUG", false), "enable debug level logging")
	log, err := newZapLogger(*debug)
	if err != nil {
		fmt.Printf("failed to initialize Zap logger: %v\n", err)
		return
	}

	flag.Parse()

	switch {
	case *help:
		flag.CommandLine.Usage()

		return
	case *version:
		return
	}

	err = utils.UpdateRLimit(log)
	if err != nil {
		log.Warning("failed to increase rlimit", zap.Error(err))
	}

	go ota.WatchUpdates(otaConfig, log)
	rand.Seed(time.Now().UnixNano())

	ctx, cancel := context.WithCancel(context.Background())
	defer cancel()

	go cancelOnSignal(cancel, log)
	job.NewRunner(runnerConfigOptions, jobsGlobalConfig).Run(ctx, log)
}

func newZapLogger(debug bool) (*zap.Logger, error) {
	var cfg zap.Config
	if debug {
		cfg = zap.NewDevelopmentConfig()

	} else {
		cfg = zap.NewProductionConfig()
	}

	cfg.EncoderConfig.EncodeTime = zapcore.TimeEncoderOfLayout("2006/01/02 15:04:05.000000")

	return cfg.Build(zap.WithCaller(false))
}

func cancelOnSignal(cancel context.CancelFunc, log *zap.Logger) {
	sigs := make(chan os.Signal, 1)
	signal.Notify(sigs,
		syscall.SIGTERM,
		syscall.SIGABRT,
		syscall.SIGHUP,
		syscall.SIGINT,
	)
	<-sigs
	log.Info("Terminating")
	cancel()
}
